@@ -1,11 +1,11 @@
 #############################################################
-# Copyright (c) 2005-2022 VMware, Inc.
+# Copyright (c) 2005-2023 VMware, Inc.
 #############################################################
 
-## @file connect.py
-## @brief Connect to a VMOMI ServiceInstance.
-##
-## Detailed description (for Doxygen goes here)
+# @file connect.py
+# @brief Connect to a VMOMI ServiceInstance.
+#
+# Detailed description (for Doxygen goes here)
 """
 Connect to a VMOMI ServiceInstance.
 
@@ -32,13 +32,14 @@ except ImportError:
     from elementtree.ElementTree import ElementTree
 
 
+TOKEN_TYPE_OAUTH_BEARER = 'oauth-bearer'
 TOKEN_TYPE_SAML = 'saml'
 TOKEN_TYPE_SSPI = 'sspi'
-TOKEN_TYPES = [TOKEN_TYPE_SAML, TOKEN_TYPE_SSPI]
+TOKEN_TYPES = [TOKEN_TYPE_OAUTH_BEARER, TOKEN_TYPE_SAML, TOKEN_TYPE_SSPI]
 
 """
 Global regular expression for parsing host and port connection
-See http://www.ietf.org/rfc/rfc3986.txt sec 3.2.2
+See https://www.ietf.org/rfc/rfc3986.txt sec 3.2.2
 """
 _rx = re.compile(r"(^\[.+\]|[^:]+)(:\d+)?$")
 
@@ -81,19 +82,19 @@ class closing(object):
 
 
 class VimSessionOrientedStub(SessionOrientedStub):
-    '''A vim-specific SessionOrientedStub.  See the SessionOrientedStub class
+    """A vim-specific SessionOrientedStub.  See the SessionOrientedStub class
     in pyVmomi/SoapAdapter.py for more information.
-    '''
+    """
 
     # The set of exceptions that should trigger a relogin by the session stub.
     SESSION_EXCEPTIONS = (vim.fault.NotAuthenticated, )
 
     @staticmethod
     def makeUserLoginMethod(username, password, locale=None):
-        '''Return a function that will call the vim.SessionManager.Login() method
+        """Return a function that will call the vim.SessionManager.Login() method
         with the given parameters.  The result of this function can be passed as
         the "loginMethod" to a SessionOrientedStub constructor.
-        '''
+        """
         def _doLogin(soapStub):
             si = vim.ServiceInstance("ServiceInstance", soapStub)
             sm = si.content.sessionManager
@@ -104,10 +105,10 @@ class VimSessionOrientedStub(SessionOrientedStub):
 
     @staticmethod
     def makeExtensionLoginMethod(extensionKey):
-        '''Return a function that will call the vim.SessionManager.Login() method
+        """Return a function that will call the vim.SessionManager.Login() method
         with the given parameters.  The result of this function can be passed as
         the "loginMethod" to a SessionOrientedStub constructor.
-        '''
+        """
         def _doLogin(soapStub):
             si = vim.ServiceInstance("ServiceInstance", soapStub)
             sm = si.content.sessionManager
@@ -118,14 +119,15 @@ class VimSessionOrientedStub(SessionOrientedStub):
         return _doLogin
 
     @staticmethod
-    def makeCertHokTokenLoginMethod(stsUrl, stsCert=None):
-        '''Return a function that will call the vim.SessionManager.LoginByToken()
+    def makeCertHokTokenLoginMethod(stsUrl, stsCert=None, ssl_context=None):
+        """Return a function that will call the vim.SessionManager.LoginByToken()
         after obtaining a HoK SAML token from the STS. The result of this function
         can be passed as the "loginMethod" to a SessionOrientedStub constructor.
 
         @param stsUrl: URL of the SAML Token issuing service. (i.e. SSO server).
         @param stsCert: public key of the STS service.
-        '''
+        @param ssl_context: SSL context
+        """
         assert (stsUrl)
 
         def _doLogin(soapStub):
@@ -135,7 +137,8 @@ class VimSessionOrientedStub(SessionOrientedStub):
             authenticator = sso.SsoAuthenticator(sts_url=stsUrl,
                                                  sts_cert=stsCert)
 
-            samlAssertion = authenticator.get_hok_saml_assertion(cert, key)
+            samlAssertion = authenticator.get_hok_saml_assertion(
+                cert, key, ssl_context=ssl_context)
 
             def _requestModifier(request):
                 return sso.add_saml_context(request, samlAssertion, key)
@@ -156,8 +159,9 @@ class VimSessionOrientedStub(SessionOrientedStub):
     def makeCredBearerTokenLoginMethod(username,
                                        password,
                                        stsUrl,
-                                       stsCert=None):
-        '''Return a function that will call the vim.SessionManager.LoginByToken()
+                                       stsCert=None,
+                                       ssl_context=None):
+        """Return a function that will call the vim.SessionManager.LoginByToken()
         after obtaining a Bearer token from the STS. The result of this function
         can be passed as the "loginMethod" to a SessionOrientedStub constructor.
 
@@ -165,7 +169,8 @@ class VimSessionOrientedStub(SessionOrientedStub):
         @param password: password of the user/service registered with STS.
         @param stsUrl: URL of the SAML Token issueing service. (i.e. SSO server).
         @param stsCert: public key of the STS service.
-        '''
+        @param ssl_context: SSL context
+        """
         assert (username)
         assert (password)
         assert (stsUrl)
@@ -177,7 +182,7 @@ class VimSessionOrientedStub(SessionOrientedStub):
             authenticator = sso.SsoAuthenticator(sts_url=stsUrl,
                                                  sts_cert=stsCert)
             samlAssertion = authenticator.get_bearer_saml_assertion(
-                username, password, cert, key)
+                username, password, cert, key, ssl_context=ssl_context)
             si = vim.ServiceInstance("ServiceInstance", soapStub)
             sm = si.content.sessionManager
             if not sm.currentSession:
@@ -208,7 +213,7 @@ def Connect(host='localhost',
             httpConnectionTimeout=None,
             connectionPoolTimeout=CONNECTION_POOL_IDLE_TIMEOUT_SEC,
             token=None,
-            tokenType=TOKEN_TYPE_SAML,
+            tokenType=None,
             disableSslCertValidation=False,
             customHeaders=None,
             # Deprecated
@@ -251,7 +256,7 @@ def Connect(host='localhost',
     @param httpProxyHost The host name of the proxy server.
     @type  httpProxyHost: string
     @param httpProxyPort The proxy server port.
-    @type  httpProxyPort: string
+    @type  httpProxyPort: int
     @param thumbprint: host cert thumbprint
     @type  thumbprint: string
     @param sslContext: SSL Context describing the various SSL options. It is only
@@ -269,7 +274,7 @@ def Connect(host='localhost',
     @param tokenType: Select which type of Authentication and Authorization token to use.
     @type  disableSslCertValidation: bool
     @param disableSslCertValidation: Creates an unverified SSL context when True.
-    @type  customHeaders: dictionary
+    @type  customHeaders: dict
     @param customHeaders: Dictionary with custom HTTP headers.
     @param b64token: base64 encoded token
            *** Deprecated: Use token instead ***
@@ -334,10 +339,10 @@ def Connect(host='localhost',
     return si
 
 
-def Disconnect(si = None):
+def Disconnect(si=None):
     """
     Logout and disconnect the service instance
-    @param si: Service instance (returned from Connect)
+    @param si: The service instance (returned from Connect)
                Defaults to the saved service instance
     """
     if not si:
@@ -358,7 +363,7 @@ def Disconnect(si = None):
         SetSi(None)
 
 
-## Method that gets a local ticket for the specified user
+# Method that gets a local ticket for the specified user
 def GetLocalTicket(si, user):
     try:
         sessionManager = si.content.sessionManager
@@ -375,8 +380,8 @@ def GetLocalTicket(si, user):
     return localTicket.userName, content
 
 
-## Private method that performs the actual Connect and returns a
-## connected service instance object.
+# Private method that performs the actual Connect and returns a
+# connected service instance object.
 
 
 def __Login(host,
@@ -426,7 +431,7 @@ def __Login(host,
     @param httpProxyHost The host name of the proxy server.
     @type  httpProxyHost: string
     @param httpProxyPort The proxy server port.
-    @type  httpProxyPort: string
+    @type  httpProxyPort: int
     @param thumbprint: host cert thumbprint
     @type  thumbprint: string
     @param sslContext: SSL Context describing the various SSL options. It is only
@@ -442,7 +447,7 @@ def __Login(host,
                     The presence of this token overrides the user and pwd parameters.
     @type  tokenType: string
     @param tokenType: Select which type of Authentication and Authorization token to use.
-    @type  customHeaders: dictionary
+    @type  customHeaders: dict
     @param customHeaders: Dictionary with custom HTTP headers.
     """
 
@@ -450,11 +455,17 @@ def __Login(host,
     if adapter != "SOAP":
         raise ValueError(adapter)
 
-    # Add an HTTP authorization header if a token is provided
-    if token and tokenType == TOKEN_TYPE_SAML:
-        if customHeaders is None:
-            customHeaders = {}
-        customHeaders.update({"Authorization": "Bearer {}".format(token)})
+    # If a token is provided
+    # add OAuth token in HTTP authorization header or
+    # add SAML token to SOAP's message wsse header
+    samlToken = None
+    if token:
+        if tokenType == TOKEN_TYPE_OAUTH_BEARER:
+            if customHeaders is None:
+                customHeaders = {}
+            customHeaders.update({"Authorization": "Bearer {}".format(token)})
+        elif tokenType == TOKEN_TYPE_SAML:
+            samlToken = token
 
     # Create the SOAP stub adapter
     stub = SoapStubAdapter(
@@ -470,6 +481,7 @@ def __Login(host,
         sslContext=sslContext,
         httpConnectionTimeout=httpConnectionTimeout,
         connectionPoolTimeout=connectionPoolTimeout,
+        samlToken=samlToken,
         customHeaders=customHeaders)
 
     # Get Service instance
@@ -504,7 +516,7 @@ def __Login(host,
     if not token:
         content.sessionManager.Login(user, pwd, None)
     else:
-        if tokenType == TOKEN_TYPE_SAML:
+        if tokenType in [TOKEN_TYPE_OAUTH_BEARER, TOKEN_TYPE_SAML]:
             content.sessionManager.LoginByToken()
         elif tokenType == TOKEN_TYPE_SSPI:
             content.sessionManager.LoginBySSPI(token)
@@ -514,7 +526,7 @@ def __Login(host,
     return si, stub
 
 
-## Get the saved service instance.
+# Get the saved service instance.
 
 
 def GetSi():
@@ -522,7 +534,7 @@ def GetSi():
     return _si
 
 
-## Set the saved service instance.
+# Set the saved service instance.
 
 
 def SetSi(si):
@@ -532,7 +544,7 @@ def SetSi(si):
     _si = si
 
 
-## Get the global saved stub
+# Get the global saved stub
 
 
 def GetStub():
@@ -543,7 +555,7 @@ def GetStub():
     return None
 
 
-## RAII-style class for managing connections
+# RAII-style class for managing connections
 
 
 class Connection(object):
@@ -578,7 +590,7 @@ class SmartConnection(object):
             self.si = None
 
 
-## Deprecated, use __GetElementTree method instead
+# Deprecated, use __GetElementTree method instead
 def __GetElementTreeFromUrl(url, sslContext):
     """
     Private method that returns an ElementTree for the XML document referenced by
@@ -611,7 +623,8 @@ def __GetElementTreeFromUrl(url, sslContext):
 
 
 def __GetElementTree(protocol, server, port, path, sslContext,
-                     httpProxyHost, httpProxyPort):
+                     httpProxyHost, httpProxyPort,
+                     customHeaders):
     """
     Private method that returns ElementTree for a remote XML document.
 
@@ -626,13 +639,20 @@ def __GetElementTree(protocol, server, port, path, sslContext,
     @param sslContext: SSL Context describing the various SSL options. It is only
                         supported in Python 2.7.9 or higher.
     @type  sslContext: SSL.Context
+    @param httpProxyHost The host name of the proxy server.
+    @type  httpProxyHost: string
+    @param httpProxyPort The proxy server port.
+    @type  httpProxyPort: int
+    @type  customHeaders: dict
+    @param customHeaders: Dictionary with custom HTTP headers.
     """
     tree = ElementTree()
+    headers = customHeaders if customHeaders else {}
 
     if httpProxyHost:
         kwargs = {"context": sslContext} if sslContext else {}
         conn = http_client.HTTPSConnection(httpProxyHost, port=httpProxyPort, **kwargs)
-        conn.set_tunnel(server, port)
+        conn.set_tunnel(server, port, headers)
     elif protocol == "https":
         kwargs = {"context": sslContext} if sslContext else {}
         conn = http_client.HTTPSConnection(server, port=port, **kwargs)
@@ -640,7 +660,7 @@ def __GetElementTree(protocol, server, port, path, sslContext,
         conn = http_client.HTTPConnection(server, port=port)
     else:
         raise Exception("Protocol " + protocol + " not supported.")
-    conn.request("GET", path)
+    conn.request(method="GET", url=path, headers=headers)
     try:
         response = conn.getresponse()
         if response.status == 200:
@@ -654,13 +674,14 @@ def __GetElementTree(protocol, server, port, path, sslContext,
         conn.close()
 
 
-## Private method that returns an ElementTree describing the API versions
-## supported by the specified server.  The result will be vimServiceVersions.xml
-## if it exists, otherwise None.
+# Private method that returns an ElementTree describing the API versions
+# supported by the specified server.  The result will be vimServiceVersions.xml
+# if it exists, otherwise None.
 
 
 def __GetServiceVersionDescription(protocol, server, port, path, sslContext,
-                                   httpProxyHost, httpProxyPort):
+                                   httpProxyHost, httpProxyPort,
+                                   customHeaders):
     """
     Private method that returns an ElementTree describing the API versions
     supported by the specified server.  The result will be vimServiceVersions.xml
@@ -677,15 +698,21 @@ def __GetServiceVersionDescription(protocol, server, port, path, sslContext,
     @param sslContext: SSL Context describing the various SSL options. It is only
                         supported in Python 2.7.9 or higher.
     @type  sslContext: SSL.Context
+    @param httpProxyHost The host name of the proxy server.
+    @type  httpProxyHost: string
+    @param httpProxyPort The proxy server port.
+    @type  httpProxyPort: int
+    @type  customHeaders: dict
+    @param customHeaders: Dictionary with custom HTTP headers.
     """
 
     return __GetElementTree(protocol, server, port,
                             path + "/vimServiceVersions.xml", sslContext,
-                            httpProxyHost, httpProxyPort)
+                            httpProxyHost, httpProxyPort, customHeaders)
 
 
-## Private method that returns true if the service version description document
-##  indicates that the desired version is supported
+# Private method that returns true if the service version description document
+#  indicates that the desired version is supported
 
 
 def __VersionIsSupported(desiredVersion, serviceVersionDescription):
@@ -721,12 +748,13 @@ def __VersionIsSupported(desiredVersion, serviceVersionDescription):
     return False
 
 
-## Private method that returns the most preferred API version supported by the
-## specified server,
+# Private method that returns the most preferred API version supported by the
+# specified server,
 
 
 def __FindSupportedVersion(protocol, server, port, path, preferredApiVersions,
-                           sslContext, httpProxyHost, httpProxyPort):
+                           sslContext, httpProxyHost, httpProxyPort,
+                           customHeaders):
     """
     Private method that returns the most preferred API version supported by the
     specified server,
@@ -746,10 +774,17 @@ def __FindSupportedVersion(protocol, server, port, path, preferredApiVersions,
     @param sslContext: SSL Context describing the various SSL options. It is only
                         supported in Python 2.7.9 or higher.
     @type  sslContext: SSL.Context
+    @param httpProxyHost The host name of the proxy server.
+    @type  httpProxyHost: string
+    @param httpProxyPort The proxy server port.
+    @type  httpProxyPort: int
+    @type  customHeaders: dict
+    @param customHeaders: Dictionary with custom HTTP headers.
     """
 
     serviceVersionDescription = __GetServiceVersionDescription(
-        protocol, server, port, path, sslContext, httpProxyHost, httpProxyPort)
+        protocol, server, port, path, sslContext, httpProxyHost,
+        httpProxyPort, customHeaders)
     if serviceVersionDescription is None:
         return None
 
@@ -781,7 +816,8 @@ def SmartStubAdapter(host='localhost',
                      sslContext=None,
                      httpConnectionTimeout=None,
                      connectionPoolTimeout=CONNECTION_POOL_IDLE_TIMEOUT_SEC,
-                     disableSslCertValidation=False):
+                     disableSslCertValidation=False,
+                     customHeaders=None):
     """
     Determine the most preferred API version supported by the specified server,
     then create a soap stub adapter using that version
@@ -795,8 +831,10 @@ def SmartStubAdapter(host='localhost',
                                     specified, the list of versions support by pyVmomi will
                                     be used.
     @type  preferredApiVersions: string or string list
-    @type disableSslCertValidation: bool
+    @type  disableSslCertValidation: bool
     @param disableSslCertValidation: Creates an unverified SSL context when True.
+    @type  customHeaders: dict
+    @param customHeaders: Dictionary with custom HTTP headers.
     """
     if preferredApiVersions is None:
         preferredApiVersions = GetServiceVersions('vim25')
@@ -806,7 +844,8 @@ def SmartStubAdapter(host='localhost',
     supportedVersion = __FindSupportedVersion('https' if port > 0 else 'http',
                                               host, port, path,
                                               preferredApiVersions, sslContext,
-                                              httpProxyHost, httpProxyPort)
+                                              httpProxyHost, httpProxyPort,
+                                              customHeaders)
     if supportedVersion is None:
         raise Exception("{0}:{1} is down or is not a VIM server"
                         .format(host, port))
@@ -829,7 +868,8 @@ def SmartStubAdapter(host='localhost',
                            samlToken=samlToken,
                            sslContext=sslContext,
                            httpConnectionTimeout=httpConnectionTimeout,
-                           connectionPoolTimeout=connectionPoolTimeout)
+                           connectionPoolTimeout=connectionPoolTimeout,
+                           customHeaders=customHeaders)
 
 
 def SmartConnect(protocol='https',
@@ -849,7 +889,7 @@ def SmartConnect(protocol='https',
                  httpConnectionTimeout=None,
                  connectionPoolTimeout=CONNECTION_POOL_IDLE_TIMEOUT_SEC,
                  token=None,
-                 tokenType=TOKEN_TYPE_SAML,
+                 tokenType=None,
                  disableSslCertValidation=False,
                  customHeaders=None,
                  # Deprecated
@@ -895,7 +935,7 @@ def SmartConnect(protocol='https',
     @param httpProxyHost The host name of the proxy server.
     @type  httpProxyHost: string
     @param httpProxyPort The proxy server port.
-    @type  httpProxyPort: string
+    @type  httpProxyPort: int
     @param thumbprint: host cert thumbprint
     @type  thumbprint: string
     @param sslContext: SSL Context describing the various SSL options. It is only
@@ -911,6 +951,8 @@ def SmartConnect(protocol='https',
                     The presence of this token overrides the user and pwd parameters.
     @type disableSslCertValidation: bool
     @param disableSslCertValidation: Creates an unverified SSL context when True.
+    @type  customHeaders: dict
+    @param customHeaders: Dictionary with custom HTTP headers.
     @param b64token: base64 encoded token
            *** Deprecated: Use token instead ***
     @type  b64token: string
@@ -926,7 +968,8 @@ def SmartConnect(protocol='https',
 
     supportedVersion = __FindSupportedVersion(protocol, host, port, path,
                                               preferredApiVersions, sslContext,
-                                              httpProxyHost, httpProxyPort)
+                                              httpProxyHost, httpProxyPort,
+                                              customHeaders)
     if supportedVersion is None:
         raise Exception("{0}:{1} is down or is not a VIM server"
                         .format(host, port))
